#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "netex.h"
#include "sockets.h"
#include "thread_manager.h"
#include "communication.h"

int server_sockfd;
struct sockaddr_in server_sockaddr;
bool b_listen = true;
char* srvName = "Linux test server";
void* srvInitBuffer = NULL;
uint32_t totalBufferSize;

int setup_server();
void* thread_listen_network();
void* thread_client_handler(void* arg);
void send_initial_pack(int client);

int start_server()
{
    PRINT_LINE("Starting server...");
    int server_result = setup_server();
    if (server_result != 0)
        return server_result;
    server_result = setup_server_thread(thread_listen_network);
    if (server_result != 0)
        return -1;
    
    char inputBuffer[100];
    
    for (;;)
    {
        if (fgets(inputBuffer, 90, stdin) != inputBuffer)
        {
            WARN("Could not get input from stdin!");
            continue;
        }
        WARN("Not a valid option given!");
        //TODO: Implement!
    }
    return 0;
}

int setup_server()
{
    server_sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (server_sockfd < 0)
    {
        ERROR("Could not create server socket!");
        return -1;
    }

    server_sockaddr.sin_family = AF_INET;
    server_sockaddr.sin_port = htons(PORT_NUM);
    server_sockaddr.sin_addr.s_addr = INADDR_ANY;

    int bind_result = bind(server_sockfd, (struct sockaddr *) &server_sockaddr, sizeof(server_sockaddr));
    if (bind_result != 0)
    {
        ERROR("Cannot bind to socket!");
        return -1;
    }
    return 0;
}

void* thread_listen_network()
{
    while (b_listen)
    {
        int listen_result = listen(server_sockfd, 1);
        if (listen_result == -1)
        {
            WARN("Could not listen. code: %i", errno);
            continue;
        }
        int clientfd = accept(server_sockfd, NULL, NULL);
        if (clientfd == -1)
        {
            WARN("Could not accept client!");
            continue;
        }
        if (init_thread(&clientfd, thread_client_handler) == -1)
            WARN("Could not establish connection with client!");
    }
}

// Function used in thread to handle the connected client.
//TODO: Return a type with status, with message on why a value is returned.
void* thread_client_handler(void* arg)
{
    int client_fd = *(int*)arg;

    // Send the initial package to the client.
    send_initial_pack(client_fd);

    char buff[50];
    // Wait for data from client.
    for(;;)
    {
        int rec_result;
        rec_result = recv(client_fd, buff, 45, 0);
        if (rec_result == -1)
        {
            WARN("Could not recieve data from client!");
            continue;
        }

        if (rec_result == 0)
        {
            PRINT_LINE("Client: %d disconnected!", client_fd);
            break;
        }
    }

    int close_result = close(client_fd);
    return NULL;
}

void send_initial_pack(int client)
{
    if (srvInitBuffer == NULL)
    {
        uint32_t srtLen = strlen(srvName) + 1;
        totalBufferSize = sizeof(uint32_t) + sizeof(EMessageType) + srtLen;
        srvInitBuffer = malloc(totalBufferSize);
        if (srvInitBuffer == NULL)
        {
            ERROR("Could not allocate memory!");
            return;
        }
        *(char*)srvInitBuffer = htonl(srtLen); // Name string lengt
        *((uint32_t*)srvInitBuffer + 1) = htonl(serverInfo); // Package type
        void* dataBuffer = ((char*)srvInitBuffer + sizeof(uint32_t) + sizeof(EMessageType)); // Push offset for string copy
        if (strcpy(dataBuffer, srvName) != dataBuffer)
        {
            ERROR("Could not copy data to buffer!");
            free(srvInitBuffer);
            srvInitBuffer = NULL;
            return;
        }
    }

    if (totalBufferSize <= 0)
        return;

    size_t totalDataSend = 0;
    while (totalDataSend < totalBufferSize)
    {
        size_t sended = send(client, srvInitBuffer, totalBufferSize - totalDataSend, 0);
        if (sended == -1)
        {
            ERROR("Could not send data.");
            break;
        }
        totalDataSend += sended;
    }
}